@@ -17,6 +17,8 @@
 
 #include "rmw_connextdds/context.hpp"
 
+#include "rmw/features.h"
+
 /*****************************************************************************
  * Context API
  *****************************************************************************/
@@ -86,6 +88,13 @@ rmw_api_connextdds_take_event(
   void * event_info,
   bool * taken);
 
+RMW_CONNEXTDDS_PUBLIC
+rmw_ret_t
+rmw_api_connextdds_event_set_callback(
+  rmw_event_t * event,
+  const rmw_event_callback_t callback,
+  const void * const user_data);
+
 /*****************************************************************************
  * Info API
  *****************************************************************************/
@@ -266,6 +275,12 @@ rmw_api_connextdds_get_gid_for_publisher(
   const rmw_publisher_t * publisher,
   rmw_gid_t * gid);
 
+RMW_CONNEXTDDS_PUBLIC
+rmw_ret_t
+rmw_api_connextdds_get_gid_for_client(
+  const rmw_client_t * client,
+  rmw_gid_t * gid);
+
 RMW_CONNEXTDDS_PUBLIC
 rmw_ret_t
 rmw_api_connextdds_compare_gids_equal(
@@ -423,6 +438,21 @@ rmw_ret_t
 rmw_api_connextdds_destroy_service(
   rmw_node_t * node,
   rmw_service_t * service);
+
+RMW_CONNEXTDDS_PUBLIC
+rmw_ret_t
+rmw_api_connextdds_service_set_on_new_request_callback(
+  rmw_service_t * rmw_service,
+  const rmw_event_callback_t callback,
+  const void * const user_data);
+
+RMW_CONNEXTDDS_PUBLIC
+rmw_ret_t
+rmw_api_connextdds_client_set_on_new_response_callback(
+  rmw_client_t * rmw_client,
+  const rmw_event_callback_t callback,
+  const void * const user_data);
+
 /*****************************************************************************
  * Subscription API
  *****************************************************************************/
@@ -459,6 +489,18 @@ rmw_api_connextdds_subscription_get_actual_qos(
   const rmw_subscription_t * subscription,
   rmw_qos_profile_t * qos);
 
+RMW_CONNEXTDDS_PUBLIC
+rmw_ret_t
+rmw_api_connextdds_subscription_set_content_filter(
+  rmw_subscription_t * subscription,
+  const rmw_subscription_content_filter_options_t * options);
+
+RMW_CONNEXTDDS_PUBLIC
+rmw_ret_t
+rmw_api_connextdds_subscription_get_content_filter(
+  const rmw_subscription_t * subscription,
+  rcutils_allocator_t * const allocator,
+  rmw_subscription_content_filter_options_t * options);
 
 RMW_CONNEXTDDS_PUBLIC
 rmw_ret_t
@@ -535,6 +577,13 @@ rmw_api_connextdds_return_loaned_message_from_subscription(
   const rmw_subscription_t * subscription,
   void * loaned_message);
 
+RMW_CONNEXTDDS_PUBLIC
+rmw_ret_t
+rmw_api_connextdds_subscription_set_on_new_message_callback(
+  rmw_subscription_t * const rmw_subscription,
+  const rmw_event_callback_t callback,
+  const void * const user_data);
+
 /*****************************************************************************
  * WaitSet API
  *****************************************************************************/
@@ -608,4 +657,11 @@ rmw_api_connextdds_subscription_get_network_flow_endpoints(
   rcutils_allocator_t * allocator,
   rmw_network_flow_endpoint_array_t * network_flow_endpoint_array);
 
+/******************************************************************************
+ * Feature support functions
+ ******************************************************************************/
+RMW_CONNEXTDDS_PUBLIC
+bool
+rmw_api_connextdds_feature_supported(rmw_feature_t feature);
+
 #endif  // RMW_CONNEXTDDS__RMW_API_IMPL_HPP_
