@@ -17,6 +17,8 @@
 #include "rmw_connextdds/discovery.hpp"
 #include "rmw_connextdds/graph_cache.hpp"
 
+#include "rosidl_runtime_c/type_hash.h"
+
 // Aliases for rmw.h functions referenced by rmw_dds_common inline functions
 #define rmw_publish     rmw_api_connextdds_publish
 
@@ -27,6 +29,7 @@ rmw_connextdds_graph_add_entityEA(
   const DDS_GUID_t * const dp_guid,
   const char * const topic_name,
   const char * const type_name,
+  const rosidl_type_hash_t & type_hash,
   const DDS_HistoryQosPolicy * const history,
   const DDS_ReliabilityQosPolicy * const reliability,
   const DDS_DurabilityQosPolicy * const durability,
@@ -727,6 +730,7 @@ rmw_connextdds_graph_add_entityEA(
   const DDS_GUID_t * const dp_guid,
   const char * const topic_name,
   const char * const type_name,
+  const rosidl_type_hash_t & type_hash,
   const DDS_HistoryQosPolicy * const history,
   const DDS_ReliabilityQosPolicy * const reliability,
   const DDS_DurabilityQosPolicy * const durability,
@@ -788,6 +792,7 @@ rmw_connextdds_graph_add_entityEA(
       gid,
       std::string(topic_name),
       std::string(type_name),
+      type_hash,
       dp_gid,
       qos_profile,
       is_reader))
@@ -875,6 +880,7 @@ rmw_connextdds_graph_add_local_publisherEA(
     &dp_guid,
     topic_name,
     pub->message_type_support()->type_name(),
+    pub->message_type_support()->type_hash(),
     &dw_qos.history,
     &dw_qos.reliability,
     &dw_qos.durability,
@@ -946,6 +952,7 @@ rmw_connextdds_graph_add_local_subscriberEA(
     &dp_guid,
     topic_name,
     sub->message_type_support()->type_name(),
+    sub->message_type_support()->type_hash(),
     &dr_qos.history,
     &dr_qos.reliability,
     &dr_qos.durability,
@@ -963,6 +970,7 @@ rmw_connextdds_graph_add_remote_entity(
   const DDS_GUID_t * const dp_guid,
   const char * const topic_name,
   const char * const type_name,
+  const DDS_UserDataQosPolicy * const user_data,
   const DDS_ReliabilityQosPolicy * const reliability,
   const DDS_DurabilityQosPolicy * const durability,
   const DDS_DeadlineQosPolicy * const deadline,
@@ -982,12 +990,28 @@ rmw_connextdds_graph_add_remote_entity(
     return RMW_RET_OK;
   }
 
+  const uint8_t * user_data_data = DDS_OctetSeq_get_contiguous_buffer(&user_data->value);
+  const size_t user_data_size = DDS_OctetSeq_get_length(&user_data->value);
+  rosidl_type_hash_t type_hash;
+  if (RMW_RET_OK != rmw_dds_common::parse_type_hash_from_user_data(
+      user_data_data, user_data_size, type_hash))
+  {
+    RMW_CONNEXT_LOG_WARNING_A(
+      "Failed to parse type hash for topic '%s' with type '%s' from USER_DATA '%*s'.",
+      topic_name, type_name,
+      static_cast<int>(user_data_size), reinterpret_cast<const char *>(user_data_data));
+    type_hash = rosidl_get_zero_initialized_type_hash();
+    // We handled the error, so clear it out
+    rmw_reset_error();
+  }
+
   rmw_ret_t rc = rmw_connextdds_graph_add_entityEA(
     ctx,
     endp_guid,
     dp_guid,
     topic_name,
     type_name,
+    type_hash,
     nullptr /* history (not propagated via discovery) */,
     reliability,
     durability,
