@@ -20,6 +20,7 @@
 #include <stdexcept>
 
 #include "rmw_dds_common/time_utils.hpp"
+#include "rmw_dds_common/qos.hpp"
 
 #include "rmw_connextdds/graph_cache.hpp"
 
@@ -29,6 +30,7 @@
 const char * const ROS_TOPIC_PREFIX = "rt";
 const char * const ROS_SERVICE_REQUESTER_PREFIX = ROS_SERVICE_REQUESTER_PREFIX_STR;
 const char * const ROS_SERVICE_RESPONSE_PREFIX = ROS_SERVICE_RESPONSE_PREFIX_STR;
+const char * const ROS_CFT_TOPIC_NAME_INFIX = "_ContentFilterTopic";
 
 std::string
 rmw_connextdds_create_topic_name(
@@ -141,7 +143,7 @@ rmw_connextdds_parse_string_list(
     input_i += 2,
     next_i_start = input_i)
   {
-    // determine token's lenght by finding a delimiter (or end of input)
+    // determine token's length by finding a delimiter (or end of input)
     for (;
       input_i + 1 < input_len && delimiter != list[input_i + 1];
       input_i += 1)
@@ -179,7 +181,7 @@ rmw_connextdds_parse_string_list(
       DDS_String_free(*el_ref);
     }
     *el_ref = DDS_String_alloc(next_len);
-    if (nullptr == el_ref) {
+    if (nullptr == *el_ref) {
       RMW_CONNEXT_LOG_ERROR_SET("failed to allocate string")
       return RMW_RET_ERROR;
     }
@@ -285,7 +287,7 @@ dds_duration_to_rmw_time(const DDS_Duration_t & duration)
   if (DDS_Duration_is_infinite(&duration)) {
     return RMW_DURATION_INFINITE;
   }
-  assert(duration.sec > 0);
+  assert(duration.sec >= 0);
   rmw_time_t result = {static_cast<uint64_t>(duration.sec), duration.nanosec};
   return result;
 }
@@ -302,12 +304,12 @@ rmw_connextdds_get_readerwriter_qos(
   DDS_ResourceLimitsQosPolicy * const resource_limits,
   DDS_PublishModeQosPolicy * const publish_mode,
   DDS_LifespanQosPolicy * const lifespan,
+  DDS_UserDataQosPolicy * const user_data,
   const rmw_qos_profile_t * const qos_policies,
   const rmw_publisher_options_t * const pub_options,
   const rmw_subscription_options_t * const sub_options)
 {
   UNUSED_ARG(writer_qos);
-  UNUSED_ARG(type_support);
   UNUSED_ARG(publish_mode);
   UNUSED_ARG(resource_limits);
   UNUSED_ARG(pub_options);
@@ -453,6 +455,21 @@ rmw_connextdds_get_readerwriter_qos(
   }
 #endif /* RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO */
 
+  std::string user_data_str;
+  if (RMW_RET_OK != rmw_dds_common::encode_type_hash_for_user_data_qos(
+      type_support->type_hash(), user_data_str))
+  {
+    RMW_CONNEXT_LOG_WARNING(
+      "Failed to encode type hash for topic, will not distribute it in USER_DATA.");
+    user_data_str.clear();
+    // We handled the error, so clear it out
+    rmw_reset_error();
+  }
+  DDS_OctetSeq_from_array(
+    &user_data->value,
+    reinterpret_cast<const uint8_t *>(user_data_str.c_str()),
+    static_cast<DDS_Long>(user_data_str.size()));
+
   return RMW_RET_OK;
 }
 
@@ -903,9 +920,11 @@ RMW_Connext_Publisher::write(
   int64_t * const sn_out)
 {
   RMW_Connext_Message user_msg;
+  if (RMW_RET_OK != RMW_Connext_Message_initialize(&user_msg, this->type_support, 0)) {
+    return RMW_RET_ERROR;
+  }
   user_msg.user_data = ros_message;
   user_msg.serialized = serialized;
-  user_msg.type_support = this->type_support;
 
   return rmw_connextdds_write_message(this, &user_msg, sn_out);
 }
@@ -947,6 +966,11 @@ rmw_ret_t
 RMW_Connext_Publisher::wait_for_all_acked(rmw_time_t wait_timeout)
 {
   DDS_Duration_t timeout = rmw_connextdds_duration_from_ros_time(&wait_timeout);
+#if RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_MICRO
+  // Avoid warnings for unused variable in micro, since wait_for_ack() is
+  // mapped to an empty call.
+  UNUSED_ARG(timeout);
+#endif  // RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_MICRO
 
   const DDS_ReturnCode_t dds_rc =
     DDS_DataWriter_wait_for_acknowledgments(this->dds_writer, &timeout);
@@ -1129,6 +1153,7 @@ RMW_Connext_Subscriber::RMW_Connext_Subscriber(
   const bool ignore_local,
   const bool created_topic,
   DDS_TopicDescription * const dds_topic_cft,
+  const char * const cft_expression,
   const bool internal)
 : internal(internal),
   ignore_local(ignore_local),
@@ -1136,6 +1161,7 @@ RMW_Connext_Subscriber::RMW_Connext_Subscriber(
   dds_reader(dds_reader),
   dds_topic(dds_topic),
   dds_topic_cft(dds_topic_cft),
+  cft_expression(cft_expression),
   type_support(type_support),
   created_topic(created_topic),
   status_condition(dds_reader, ignore_local, internal)
@@ -1236,7 +1262,7 @@ RMW_Connext_Subscriber::create(
   }
 
   auto scope_exit_topic_delete = rcpputils::make_scope_exit(
-    [ctx, topic_created, dp, topic, cft_topic]()
+    [ctx, &topic_created, dp, &topic, &cft_topic]()
     {
       if (nullptr != cft_topic) {
         if (RMW_RET_OK !=
@@ -1256,21 +1282,42 @@ RMW_Connext_Subscriber::create(
     });
 
   DDS_TopicDescription * sub_topic = DDS_Topic_as_topicdescription(topic);
+  std::string sub_cft_name;
+  const char * sub_cft_expr = "";
+  const rcutils_string_array_t * sub_cft_params = nullptr;
 
   if (nullptr != cft_name) {
-    rmw_ret_t cft_rc =
-      rmw_connextdds_create_contentfilteredtopic(
-      ctx, dp, topic, cft_name, cft_filter, &cft_topic);
-
-    if (RMW_RET_OK != cft_rc) {
-      if (RMW_RET_UNSUPPORTED != cft_rc) {
-        return nullptr;
-      }
-    } else {
-      sub_topic = cft_topic;
+    sub_cft_name = cft_name;
+    sub_cft_expr = cft_filter;
+  } else {
+    sub_cft_name =
+      fqtopic_name + ROS_CFT_TOPIC_NAME_INFIX + RMW_Connext_Subscriber::get_atomic_id();
+    if (nullptr != subscriber_options->content_filter_options) {
+      sub_cft_expr =
+        subscriber_options->content_filter_options->filter_expression;
+      sub_cft_params =
+        &subscriber_options->content_filter_options->expression_parameters;
     }
   }
 
+  rmw_ret_t cft_rc =
+    rmw_connextdds_create_contentfilteredtopic(
+    ctx,
+    dp,
+    topic,
+    sub_cft_name.c_str(),
+    sub_cft_expr,
+    sub_cft_params,
+    &cft_topic);
+
+  if (RMW_RET_OK != cft_rc) {
+    if (RMW_RET_UNSUPPORTED != cft_rc) {
+      return nullptr;
+    }
+  } else {
+    sub_topic = cft_topic;
+  }
+
   // The following initialization generates warnings when built
   // with RTI Connext DDS Professional < 6 (e.g. 5.3.1), so use
   // DDS_DataWriterQos_initialize() for older versions.
@@ -1337,15 +1384,16 @@ RMW_Connext_Subscriber::create(
     subscriber_options->ignore_local_publications,
     topic_created,
     cft_topic,
+    sub_cft_expr,
     internal);
 
   if (nullptr == rmw_sub_impl) {
     RMW_CONNEXT_LOG_ERROR_SET("failed to allocate RMW subscriber")
     return nullptr;
   }
-  scope_exit_type_unregister.cancel();
-  scope_exit_topic_delete.cancel();
   scope_exit_dds_reader_delete.cancel();
+  scope_exit_topic_delete.cancel();
+  scope_exit_type_unregister.cancel();
 
   return rmw_sub_impl;
 }
@@ -1529,6 +1577,30 @@ RMW_Connext_Subscriber::take_serialized(
   return rc;
 }
 
+
+rmw_ret_t
+RMW_Connext_Subscriber::set_content_filter(
+  const rmw_subscription_content_filter_options_t * const options)
+{
+  if (RMW_RET_OK !=
+    rmw_connextdds_set_cft_filter_expression(
+      this->dds_topic_cft, options->filter_expression, &options->expression_parameters))
+  {
+    return RMW_RET_ERROR;
+  }
+
+  this->cft_expression = options->filter_expression;
+  return RMW_RET_OK;
+}
+
+rmw_ret_t
+RMW_Connext_Subscriber::get_content_filter(
+  rcutils_allocator_t * allocator,
+  rmw_subscription_content_filter_options_t * const options)
+{
+  return rmw_connextdds_get_cft_filter_expression(this->dds_topic_cft, allocator, options);
+}
+
 rmw_ret_t
 RMW_Connext_Subscriber::loan_messages(const bool update_condition)
 {
@@ -1634,8 +1706,8 @@ RMW_Connext_Subscriber::take_next(
       RMW_RET_OK == rc_exit;
       this->loan_next++)
     {
-      rcutils_uint8_array_t * data_buffer =
-        reinterpret_cast<rcutils_uint8_array_t *>(
+      RMW_Connext_Message * msg =
+        reinterpret_cast<RMW_Connext_Message *>(
         DDS_UntypedSampleSeq_get_reference(
           &this->loan_data, static_cast<DDS_Long>(this->loan_next)));
       DDS_SampleInfo * info =
@@ -1656,7 +1728,7 @@ RMW_Connext_Subscriber::take_next(
 
             if (RMW_RET_OK !=
               this->type_support->deserialize(
-                ros_message, data_buffer, deserialized_size, true /* header_only */))
+                ros_message, &msg->data_buffer, deserialized_size, true /* header_only */))
             {
               RMW_CONNEXT_LOG_ERROR_SET("failed to deserialize taken sample")
               rc_exit = RMW_RET_ERROR;
@@ -1695,7 +1767,7 @@ RMW_Connext_Subscriber::take_next(
           if (RCUTILS_RET_OK !=
             rcutils_uint8_array_copy(
               reinterpret_cast<rcutils_uint8_array_t *>(ros_message),
-              data_buffer))
+              &msg->data_buffer))
           {
             RMW_CONNEXT_LOG_ERROR_SET("failed to copy uint8 array")
             rc_exit = RMW_RET_ERROR;
@@ -1706,7 +1778,7 @@ RMW_Connext_Subscriber::take_next(
 
           if (RMW_RET_OK !=
             this->type_support->deserialize(
-              ros_message, data_buffer, deserialized_size))
+              ros_message, &msg->data_buffer, deserialized_size))
           {
             RMW_CONNEXT_LOG_ERROR_SET(
               "failed to deserialize taken sample")
@@ -1771,7 +1843,6 @@ rmw_connextdds_create_subscriber(
       "failed to allocate RMW_Connext_Subscriber")
     return nullptr;
   }
-
   auto scope_exit_rmw_reader_impl_delete =
     rcpputils::make_scope_exit(
     [rmw_sub_impl]()
@@ -1782,6 +1853,25 @@ rmw_connextdds_create_subscriber(
       }
       delete rmw_sub_impl;
     });
+#if RMW_CONNEXT_DEBUG && RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO
+  auto scope_exit_enable_participant_on_error =
+    rcpputils::make_scope_exit(
+    [ctx]()
+    {
+      // If we are building in Debug mode, an issue in Connext may prevent the
+      // participant from being able to delete any content-filtered topic if
+      // the participant has not been enabled.
+      // For this reason, make sure to enable the participant before trying to
+      // finalize it.
+      // TODO(asorbini) reconsider the need for this code in Connext > 6.1.0
+      if (DDS_RETCODE_OK !=
+      DDS_Entity_enable(DDS_DomainParticipant_as_entity(ctx->participant)))
+      {
+        RMW_CONNEXT_LOG_ERROR_SET(
+          "failed to enable DomainParticipant on subscriber creation error")
+      }
+    });
+#endif  // RMW_CONNEXT_DEBUG && RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO
 
   rmw_subscription_t * rmw_subscriber = rmw_subscription_allocate();
   if (nullptr == rmw_subscriber) {
@@ -1815,6 +1905,7 @@ rmw_connextdds_create_subscriber(
     topic_name_len + 1);
   rmw_subscriber->options = *subscriber_options;
   rmw_subscriber->can_loan_messages = false;
+  rmw_subscriber->is_cft_enabled = rmw_sub_impl->is_cft_enabled();
 
   if (!internal) {
     if (RMW_RET_OK != rmw_sub_impl->enable()) {
@@ -1831,6 +1922,9 @@ rmw_connextdds_create_subscriber(
     }
   }
 
+#if RMW_CONNEXT_DEBUG && RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO
+  scope_exit_enable_participant_on_error.cancel();
+#endif  // RMW_CONNEXT_DEBUG && RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO
   scope_exit_rmw_reader_impl_delete.cancel();
   scope_exit_rmw_reader_delete.cancel();
   return rmw_subscriber;
@@ -1885,6 +1979,20 @@ rmw_connextdds_message_info_from_dds(
   to->source_timestamp = 0;
   to->received_timestamp = 0;
 #endif /* !RTI_WIN32 */
+  // Currently we cannot use `rmw_connextdds_sn_dds_to_ros`, as that was used to convert to
+  // `rmw_request_id_t.sequence_number`, which is an int64_t and not an uint64_t.
+  // When rmw is updated and all sequence numbers are a `uint64_t`,
+  // rmw_connextdds_sn_dds_to_ros() should be updated and used everywhere.
+  to->publication_sequence_number =
+    static_cast<uint64_t>((from->publication_sequence_number).high) << 32 |
+    static_cast<uint64_t>((from->publication_sequence_number).low);
+#if RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO
+  to->reception_sequence_number =
+    static_cast<uint64_t>((from->reception_sequence_number).high) << 32 |
+    static_cast<uint64_t>((from->reception_sequence_number).low);
+#else
+  to->reception_sequence_number = RMW_MESSAGE_INFO_SEQUENCE_NUMBER_UNSUPPORTED;
+#endif  // RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO
 }
 
 /******************************************************************************
@@ -2470,6 +2578,12 @@ RMW_Connext_Client::enable()
 rmw_ret_t
 RMW_Connext_Client::is_service_available(bool & available)
 {
+#if RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_MICRO
+  available = 0 < this->request_pub->subscriptions_count() &&
+    0 < this->reply_sub->publications_count();
+  return RMW_RET_OK;
+#else /* RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO */
+
   // mark service as available if we have at least one writer and one reader
   // matched from the same remote DomainParticipant.
   struct DDS_InstanceHandleSeq matched_req_subs = DDS_SEQUENCE_INITIALIZER,
@@ -2514,6 +2628,7 @@ RMW_Connext_Client::is_service_available(bool & available)
   }
 
   return RMW_RET_OK;
+#endif /* RMW_CONNEXT_DDS_API == RMW_CONNEXT_DDS_API_PRO */
 }
 
 rmw_ret_t
@@ -2982,6 +3097,19 @@ ros_event_to_dds(const rmw_event_type_t ros, bool * const invalid)
       {
         return DDS_SAMPLE_LOST_STATUS;
       }
+    case RMW_EVENT_PUBLISHER_INCOMPATIBLE_TYPE:
+    case RMW_EVENT_SUBSCRIPTION_INCOMPATIBLE_TYPE:
+      {
+        return DDS_INCONSISTENT_TOPIC_STATUS;
+      }
+    case RMW_EVENT_PUBLICATION_MATCHED:
+      {
+        return DDS_PUBLICATION_MATCHED_STATUS;
+      }
+    case RMW_EVENT_SUBSCRIPTION_MATCHED:
+      {
+        return DDS_SUBSCRIPTION_MATCHED_STATUS;
+      }
     default:
       {
         if (nullptr != invalid) {
@@ -3024,6 +3152,10 @@ dds_event_to_str(const DDS_StatusKind event)
       {
         return "SAMPLE_LOST";
       }
+    case DDS_INCONSISTENT_TOPIC_STATUS:
+      {
+        return "INCONSISTENT_TOPIC";
+      }
     default:
       {
         return "UNSUPPORTED";
@@ -3039,6 +3171,8 @@ ros_event_for_reader(const rmw_event_type_t ros)
     case RMW_EVENT_REQUESTED_DEADLINE_MISSED:
     case RMW_EVENT_REQUESTED_QOS_INCOMPATIBLE:
     case RMW_EVENT_MESSAGE_LOST:
+    case RMW_EVENT_SUBSCRIPTION_INCOMPATIBLE_TYPE:
+    case RMW_EVENT_SUBSCRIPTION_MATCHED:
       {
         return true;
       }
@@ -3088,6 +3222,22 @@ RMW_Connext_SubscriberStatusCondition::get_status(
         rc = this->get_message_lost_status(status);
         break;
       }
+    case RMW_EVENT_SUBSCRIPTION_INCOMPATIBLE_TYPE:
+      {
+        rmw_incompatible_type_status_t * const status =
+          reinterpret_cast<rmw_incompatible_type_status_t *>(event_info);
+
+        rc = this->get_incompatible_type_status(status);
+        break;
+      }
+    case RMW_EVENT_SUBSCRIPTION_MATCHED:
+      {
+        rmw_matched_status_t * const status =
+          reinterpret_cast<rmw_matched_status_t *>(event_info);
+
+        rc = this->get_matched_status(status);
+        break;
+      }
     default:
       {
         RMW_CONNEXT_LOG_ERROR_A_SET(
@@ -3131,6 +3281,22 @@ RMW_Connext_PublisherStatusCondition::get_status(
         rc = this->get_offered_qos_incompatible_status(status);
         break;
       }
+    case RMW_EVENT_PUBLISHER_INCOMPATIBLE_TYPE:
+      {
+        rmw_incompatible_type_status_t * const status =
+          reinterpret_cast<rmw_incompatible_type_status_t *>(event_info);
+
+        rc = this->get_incompatible_type_status(status);
+        break;
+      }
+    case RMW_EVENT_PUBLICATION_MATCHED:
+      {
+        rmw_matched_status_t * status =
+          reinterpret_cast<rmw_matched_status_t *>(event_info);
+
+        rc = this->get_matched_status(status);
+        break;
+      }
     default:
       {
         RMW_CONNEXT_LOG_ERROR_A_SET("unsupported publisher qos: %d", event_type)
