@@ -111,6 +111,20 @@ RMW_Connext_DataReaderListener_sample_lost(
   self->on_sample_lost(status);
 }
 
+void
+RMW_Connext_DataReaderListener_matched(
+  void * listener_data,
+  DDS_DataReader * reader,
+  const struct DDS_SubscriptionMatchedStatus * status)
+{
+  RMW_Connext_SubscriberStatusCondition * const self =
+    reinterpret_cast<RMW_Connext_SubscriberStatusCondition *>(listener_data);
+
+  UNUSED_ARG(reader);
+
+  self->on_matched(status);
+}
+
 void
 RMW_Connext_DataReaderListener_on_data_available(
   void * listener_data,
@@ -121,6 +135,7 @@ RMW_Connext_DataReaderListener_on_data_available(
 
   UNUSED_ARG(reader);
 
+  self->notify_new_data();
   self->set_data_available(true);
 }
 
@@ -166,6 +181,33 @@ RMW_Connext_DataWriterListener_liveliness_lost(
   self->on_liveliness_lost(status);
 }
 
+void
+RMW_Connext_TopicListener_on_inconsistent_topic(
+  void * listener_data,
+  DDS_Topic * topic,
+  const struct DDS_InconsistentTopicStatus * status)
+{
+  RMW_Connext_StatusCondition * const self =
+    reinterpret_cast<RMW_Connext_StatusCondition *>(listener_data);
+
+  UNUSED_ARG(topic);
+
+  self->on_inconsistent_topic(status);
+}
+
+void
+RMW_Connext_DataWriterListener_matched(
+  void * listener_data,
+  DDS_DataWriter * writer,
+  const struct DDS_PublicationMatchedStatus * status)
+{
+  RMW_Connext_PublisherStatusCondition * const self =
+    reinterpret_cast<RMW_Connext_PublisherStatusCondition *>(listener_data);
+
+  UNUSED_ARG(writer);
+
+  self->on_matched(status);
+}
 
 bool
 RMW_Connext_WaitSet::on_condition_active(
@@ -550,6 +592,24 @@ RMW_Connext_WaitSet::wait(
   return RMW_RET_OK;
 }
 
+void
+RMW_Connext_StatusCondition::on_inconsistent_topic(
+  const struct DDS_InconsistentTopicStatus * status)
+{
+  update_state(
+    [this, status]() {
+      this->update_status_inconsistent_topic(status);
+    }, true /* notify */);
+}
+
+void
+RMW_Connext_StatusCondition::update_status_inconsistent_topic(
+  const struct DDS_InconsistentTopicStatus * status)
+{
+  this->status_inconsistent_topic = *status;
+  this->triggered_inconsistent_topic = true;
+}
+
 rmw_ret_t
 RMW_Connext_SubscriberStatusCondition::install(
   RMW_Connext_Subscriber * const sub)
@@ -567,6 +627,8 @@ RMW_Connext_SubscriberStatusCondition::install(
     RMW_Connext_DataReaderListener_sample_lost;
   listener.on_data_available =
     RMW_Connext_DataReaderListener_on_data_available;
+  listener.on_subscription_matched =
+    RMW_Connext_DataReaderListener_matched;
   listener.as_listener.listener_data = this;
 
   listener_mask =
@@ -574,6 +636,7 @@ RMW_Connext_SubscriberStatusCondition::install(
     DDS_REQUESTED_INCOMPATIBLE_QOS_STATUS |
     DDS_LIVELINESS_CHANGED_STATUS |
     DDS_SAMPLE_LOST_STATUS |
+    DDS_SUBSCRIPTION_MATCHED_STATUS |
     DDS_DATA_AVAILABLE_STATUS;
 
   rmw_connextdds_configure_subscriber_condition_listener(
@@ -586,6 +649,17 @@ RMW_Connext_SubscriberStatusCondition::install(
     return RMW_RET_ERROR;
   }
 
+  struct DDS_TopicListener topic_listener = DDS_TopicListener_INITIALIZER;
+  topic_listener.on_inconsistent_topic = RMW_Connext_TopicListener_on_inconsistent_topic;
+  topic_listener.as_listener.listener_data = this;
+
+  if (DDS_RETCODE_OK !=
+    DDS_Topic_set_listener(sub->topic(), &topic_listener, DDS_INCONSISTENT_TOPIC_STATUS))
+  {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to set topic listener");
+    return RMW_RET_ERROR;
+  }
+
   this->sub = sub;
 
   return RMW_RET_OK;
@@ -607,10 +681,12 @@ RMW_Connext_SubscriberStatusCondition::RMW_Connext_SubscriberStatusCondition(
   status_qos(DDS_RequestedIncompatibleQosStatus_INITIALIZER),
   status_liveliness(DDS_LivelinessChangedStatus_INITIALIZER),
   status_sample_lost(DDS_SampleLostStatus_INITIALIZER),
+  status_matched(DDS_SubscriptionMatchedStatus_INITIALIZER),
   status_deadline_last(DDS_RequestedDeadlineMissedStatus_INITIALIZER),
   status_qos_last(DDS_RequestedIncompatibleQosStatus_INITIALIZER),
   status_liveliness_last(DDS_LivelinessChangedStatus_INITIALIZER),
   status_sample_lost_last(DDS_SampleLostStatus_INITIALIZER),
+  status_matched_last(DDS_SubscriptionMatchedStatus_INITIALIZER),
   sub(nullptr)
 {
   if (internal && nullptr == this->loan_guard_condition) {
@@ -649,6 +725,14 @@ RMW_Connext_SubscriberStatusCondition::has_status(
       {
         return this->triggered_sample_lost;
       }
+    case RMW_EVENT_SUBSCRIPTION_INCOMPATIBLE_TYPE:
+      {
+        return this->triggered_inconsistent_topic;
+      }
+    case RMW_EVENT_SUBSCRIPTION_MATCHED:
+      {
+        return this->triggered_matched;
+      }
     default:
       {
         RMW_CONNEXT_ASSERT(0)
@@ -697,6 +781,16 @@ RMW_Connext_SubscriberStatusCondition::on_sample_lost(
     }, true /* notify */);
 }
 
+void
+RMW_Connext_SubscriberStatusCondition::on_matched(
+  const DDS_SubscriptionMatchedStatus * const status)
+{
+  update_state(
+    [this, status]() {
+      this->update_status_matched(status);
+    }, true /* notify */);
+}
+
 void
 RMW_Connext_SubscriberStatusCondition::update_status_deadline(
   const DDS_RequestedDeadlineMissedStatus * const status)
@@ -706,6 +800,8 @@ RMW_Connext_SubscriberStatusCondition::update_status_deadline(
 
   this->status_deadline.total_count_change = this->status_deadline.total_count;
   this->status_deadline.total_count_change -= this->status_deadline_last.total_count;
+
+  this->notify_new_event(RMW_EVENT_REQUESTED_DEADLINE_MISSED);
 }
 
 void
@@ -720,6 +816,8 @@ RMW_Connext_SubscriberStatusCondition::update_status_liveliness(
   this->status_liveliness.alive_count_change -= this->status_liveliness_last.alive_count;
   this->status_liveliness.not_alive_count_change -=
     this->status_liveliness_last.not_alive_count;
+
+  this->notify_new_event(RMW_EVENT_LIVELINESS_CHANGED);
 }
 
 void
@@ -731,6 +829,7 @@ RMW_Connext_SubscriberStatusCondition::update_status_qos(
 
   this->status_qos.total_count_change = this->status_qos.total_count;
   this->status_qos.total_count_change -= this->status_qos_last.total_count;
+  this->notify_new_event(RMW_EVENT_REQUESTED_QOS_INCOMPATIBLE);
 }
 
 void
@@ -743,6 +842,46 @@ RMW_Connext_SubscriberStatusCondition::update_status_sample_lost(
   this->status_sample_lost.total_count_change = this->status_sample_lost.total_count;
   this->status_sample_lost.total_count_change -=
     this->status_sample_lost_last.total_count;
+  this->notify_new_event(RMW_EVENT_MESSAGE_LOST);
+}
+
+void
+RMW_Connext_SubscriberStatusCondition::notify_new_data()
+{
+  size_t unread_samples = 0;
+  std::unique_lock<std::mutex> lock_mutex(new_data_event_mutex_);
+  perform_action_and_update_state(
+    [this, &unread_samples]() {
+      const rmw_ret_t rc = this->sub->count_unread_samples(unread_samples);
+      if (RMW_RET_OK != rc) {
+        RMW_CONNEXT_LOG_ERROR("failed to count unread samples on DDS Reader")
+      }
+    },
+    [this, &unread_samples]() {
+      if (unread_samples == 0) {
+        return;
+      }
+      if (new_data_event_cb_) {
+        new_data_event_cb_(data_event_user_data_, unread_samples);
+      } else {
+        unread_data_events_count_ += unread_samples;
+      }
+    });
+}
+
+void
+RMW_Connext_SubscriberStatusCondition::update_status_matched(
+  const DDS_SubscriptionMatchedStatus * const status)
+{
+  this->status_matched = *status;
+  this->triggered_matched = true;
+
+  this->status_matched.total_count_change =
+    this->status_matched.total_count - this->status_matched_last.total_count;
+  this->status_matched.current_count_change =
+    this->status_matched.current_count - this->status_matched_last.current_count;
+
+  this->notify_new_event(RMW_EVENT_SUBSCRIPTION_MATCHED);
 }
 
 rmw_ret_t
@@ -758,12 +897,15 @@ RMW_Connext_PublisherStatusCondition::install(
     RMW_Connext_DataWriterListener_offered_incompatible_qos;
   listener.on_liveliness_lost =
     RMW_Connext_DataWriterListener_liveliness_lost;
+  listener.on_publication_matched =
+    RMW_Connext_DataWriterListener_matched;
   listener.as_listener.listener_data = this;
 
   listener_mask =
     DDS_OFFERED_DEADLINE_MISSED_STATUS |
     DDS_OFFERED_INCOMPATIBLE_QOS_STATUS |
-    DDS_LIVELINESS_LOST_STATUS;
+    DDS_LIVELINESS_LOST_STATUS |
+    DDS_PUBLICATION_MATCHED_STATUS;
 
   if (DDS_RETCODE_OK !=
     DDS_DataWriter_set_listener(
@@ -773,6 +915,23 @@ RMW_Connext_PublisherStatusCondition::install(
     return RMW_RET_ERROR;
   }
 
+  DDS_Topic * topic = DDS_DataWriter_get_topic(pub->writer());
+  if (topic == nullptr) {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to get topic associated with data writer");
+    return RMW_RET_ERROR;
+  }
+
+  struct DDS_TopicListener topic_listener = DDS_TopicListener_INITIALIZER;
+  topic_listener.on_inconsistent_topic = RMW_Connext_TopicListener_on_inconsistent_topic;
+  topic_listener.as_listener.listener_data = this;
+
+  if (DDS_RETCODE_OK !=
+    DDS_Topic_set_listener(topic, &topic_listener, DDS_INCONSISTENT_TOPIC_STATUS))
+  {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to set topic listener");
+    return RMW_RET_ERROR;
+  }
+
   this->pub = pub;
 
   return RMW_RET_OK;
@@ -784,9 +943,11 @@ RMW_Connext_PublisherStatusCondition::RMW_Connext_PublisherStatusCondition(
   status_deadline(DDS_OfferedDeadlineMissedStatus_INITIALIZER),
   status_qos(DDS_OfferedIncompatibleQosStatus_INITIALIZER),
   status_liveliness(DDS_LivelinessLostStatus_INITIALIZER),
+  status_matched(DDS_PublicationMatchedStatus_INITIALIZER),
   status_deadline_last(DDS_OfferedDeadlineMissedStatus_INITIALIZER),
   status_qos_last(DDS_OfferedIncompatibleQosStatus_INITIALIZER),
-  status_liveliness_last(DDS_LivelinessLostStatus_INITIALIZER)
+  status_liveliness_last(DDS_LivelinessLostStatus_INITIALIZER),
+  status_matched_last(DDS_PublicationMatchedStatus_INITIALIZER)
 {}
 
 bool
@@ -806,6 +967,14 @@ RMW_Connext_PublisherStatusCondition::has_status(
       {
         return this->triggered_qos;
       }
+    case RMW_EVENT_PUBLISHER_INCOMPATIBLE_TYPE:
+      {
+        return this->triggered_inconsistent_topic;
+      }
+    case RMW_EVENT_PUBLICATION_MATCHED:
+      {
+        return this->triggered_matched;
+      }
     default:
       RMW_CONNEXT_ASSERT(0)
       return false;
@@ -843,6 +1012,16 @@ RMW_Connext_PublisherStatusCondition::on_liveliness_lost(
     }, true /* notify */);
 }
 
+void
+RMW_Connext_PublisherStatusCondition::on_matched(
+  const DDS_PublicationMatchedStatus * const status)
+{
+  update_state(
+    [this, status]() {
+      this->update_status_matched(status);
+    }, true /* notify */);
+}
+
 void
 RMW_Connext_PublisherStatusCondition::update_status_deadline(
   const DDS_OfferedDeadlineMissedStatus * const status)
@@ -852,6 +1031,8 @@ RMW_Connext_PublisherStatusCondition::update_status_deadline(
 
   this->status_deadline.total_count_change = this->status_deadline.total_count;
   this->status_deadline.total_count_change -= this->status_deadline_last.total_count;
+
+  this->notify_new_event(RMW_EVENT_OFFERED_DEADLINE_MISSED);
 }
 
 void
@@ -863,6 +1044,8 @@ RMW_Connext_PublisherStatusCondition::update_status_liveliness(
 
   this->status_liveliness.total_count_change = this->status_liveliness.total_count;
   this->status_liveliness.total_count_change -= this->status_liveliness_last.total_count;
+
+  this->notify_new_event(RMW_EVENT_LIVELINESS_CHANGED);
 }
 
 void
@@ -874,4 +1057,21 @@ RMW_Connext_PublisherStatusCondition::update_status_qos(
 
   this->status_qos.total_count_change = this->status_qos.total_count;
   this->status_qos.total_count_change -= this->status_qos_last.total_count;
+
+  this->notify_new_event(RMW_EVENT_OFFERED_QOS_INCOMPATIBLE);
+}
+
+void
+RMW_Connext_PublisherStatusCondition::update_status_matched(
+  const DDS_PublicationMatchedStatus * const status)
+{
+  this->status_matched = *status;
+  this->triggered_matched = true;
+
+  this->status_matched.total_count_change =
+    this->status_matched.total_count - this->status_matched_last.total_count;
+  this->status_matched.current_count_change =
+    this->status_matched.current_count - this->status_matched_last.current_count;
+
+  this->notify_new_event(RMW_EVENT_PUBLICATION_MATCHED);
 }
