@@ -15,8 +15,10 @@
 #include <string>
 #include <map>
 #include <vector>
+#include <cmath>
 
 #include "rmw/impl/cpp/key_value.hpp"
+#include "rmw_connextdds/custom_sql_filter.hpp"
 
 #include "rmw_connextdds/type_support.hpp"
 #include "rmw_connextdds/rmw_impl.hpp"
@@ -25,6 +27,13 @@
 const char * const RMW_CONNEXTDDS_ID = "rmw_connextdds";
 const char * const RMW_CONNEXTDDS_SERIALIZATION_FORMAT = "cdr";
 
+struct rmw_connextdds_api_pro
+{
+  rti_connext_dds_custom_sql_filter::CustomSqlFilterData custom_filter_data;
+};
+
+rmw_connextdds_api_pro * RMW_Connext_fv_FactoryContext = nullptr;
+
 rmw_ret_t
 rmw_connextdds_set_log_verbosity(rmw_log_severity_t severity)
 {
@@ -69,9 +78,28 @@ rmw_ret_t
 rmw_connextdds_initialize_participant_factory_context(
   rmw_context_impl_t * const ctx)
 {
+  RMW_CONNEXT_ASSERT(RMW_Connext_fv_FactoryContext == nullptr)
+  // RMW_Connext_gv_DomainParticipantFactory is initialized by
+  // rmw_api_connextdds_init().
   RMW_CONNEXT_ASSERT(RMW_Connext_gv_DomainParticipantFactory == nullptr)
   UNUSED_ARG(ctx);
-  // Nothing to do
+
+  rmw_connextdds_api_pro * ctx_api = nullptr;
+  auto scope_exit_api_delete = rcpputils::make_scope_exit(
+    [ctx_api]()
+    {
+      if (nullptr != ctx_api) {
+        delete ctx_api;
+      }
+    });
+
+  ctx_api = new (std::nothrow) rmw_connextdds_api_pro();
+  if (nullptr == ctx_api) {
+    return RMW_RET_ERROR;
+  }
+
+  scope_exit_api_delete.cancel();
+  RMW_Connext_fv_FactoryContext = ctx_api;
   return RMW_RET_OK;
 }
 
@@ -79,6 +107,71 @@ rmw_ret_t
 rmw_connextdds_finalize_participant_factory_context(
   rmw_context_impl_t * const ctx)
 {
+  RMW_CONNEXT_ASSERT(nullptr != RMW_Connext_fv_FactoryContext)
+  rmw_connextdds_api_pro * const ctx_api = RMW_Connext_fv_FactoryContext;
+  RMW_Connext_fv_FactoryContext = nullptr;
+
+  delete ctx_api;
+
+  if (nullptr == RMW_Connext_gv_DomainParticipantFactory) {
+    // Nothing else  to do if we didn't even initialize the factory;
+    return RMW_RET_OK;
+  }
+
+  // There might be some DomainParticipants left-over from a ("failed context
+  // initialization" + "failed participant finalization"), so let's try to
+  // clean them up.
+  DDS_DomainParticipantSeq participants = DDS_SEQUENCE_INITIALIZER;
+  auto scope_exit_seq = rcpputils::make_scope_exit(
+    [&participants]()
+    {
+      DDS_DomainParticipantSeq_finalize(&participants);
+    });
+
+  if (DDS_RETCODE_OK !=
+    DDS_DomainParticipantFactory_get_participants(
+      RMW_Connext_gv_DomainParticipantFactory, &participants))
+  {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to list existing participants")
+    return RMW_RET_ERROR;
+  }
+
+  const DDS_Long pending = DDS_DomainParticipantSeq_get_length(&participants);
+  for (DDS_Long i = 0; i < pending; i++) {
+    DDS_DomainParticipant * const participant =
+      *DDS_DomainParticipantSeq_get_reference(&participants, i);
+#if RMW_CONNEXT_DEBUG
+    // If we are building in Debug mode, an issue in Connext may prevent the
+    // participant from being able to delete any content-filtered topic if
+    // the participant has not been enabled.
+    // For this reason, make sure to enable the participant before trying to
+    // finalize it.
+    // TODO(asorbini) reconsider the need for this code in Connext > 6.1.0
+    if (DDS_RETCODE_OK !=
+      DDS_Entity_enable(DDS_DomainParticipant_as_entity(participant)))
+    {
+      RMW_CONNEXT_LOG_ERROR_SET(
+        "failed to enable pending DomainParticipant before deletion")
+      return RMW_RET_ERROR;
+    }
+#endif  // RMW_CONNEXT_DEBUG
+    if (DDS_RETCODE_OK !=
+      DDS_DomainParticipant_delete_contained_entities(participant))
+    {
+      RMW_CONNEXT_LOG_ERROR_SET(
+        "failed to delete pending DomainParticipant's entities")
+      return RMW_RET_ERROR;
+    }
+
+    if (DDS_RETCODE_OK !=
+      DDS_DomainParticipantFactory_delete_participant(
+        RMW_Connext_gv_DomainParticipantFactory, participant))
+    {
+      RMW_CONNEXT_LOG_ERROR_SET("failed to delete pending DomainParticipant")
+      return RMW_RET_ERROR;
+    }
+  }
+
   UNUSED_ARG(ctx);
   return RMW_RET_OK;
 }
@@ -110,23 +203,6 @@ rmw_connextdds_initialize_participant_qos_impl(
     case rmw_context_impl_t::participant_qos_override_policy_t::All:
     case rmw_context_impl_t::participant_qos_override_policy_t::Basic:
       {
-        // Parse and apply QoS parameters derived from ROS 2 configuration options.
-
-        if (ctx->localhost_only) {
-          if (DDS_RETCODE_OK !=
-            DDS_PropertyQosPolicyHelper_assert_property(
-              &dp_qos->property,
-              "dds.transport.UDPv4.builtin.parent.allow_interfaces",
-              RMW_CONNEXT_LOCALHOST_ONLY_ADDRESS,
-              DDS_BOOLEAN_FALSE /* propagate */))
-          {
-            RMW_CONNEXT_LOG_ERROR_A_SET(
-              "failed to assert property on participant: %s",
-              "dds.transport.UDPv4.builtin.parent.allow_interfaces")
-            return RMW_RET_ERROR;
-          }
-        }
-
         const size_t user_data_len_in =
           DDS_OctetSeq_get_length(&dp_qos->user_data.value);
 
@@ -229,6 +305,48 @@ rmw_connextdds_initialize_participant_qos_impl(
   return RMW_RET_OK;
 }
 
+rmw_ret_t
+rmw_connextdds_configure_participant(
+  rmw_context_impl_t * const ctx,
+  DDS_DomainParticipant * const participant)
+{
+  UNUSED_ARG(ctx);
+
+  if (DDS_RETCODE_OK !=
+    rti_connext_dds_custom_sql_filter::register_content_filter(
+      participant, &RMW_Connext_fv_FactoryContext->custom_filter_data))
+  {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to register custom SQL filter")
+    return RMW_RET_ERROR;
+  }
+
+  return RMW_RET_OK;
+}
+
+static
+rmw_ret_t
+rmw_connextdds_initialize_cft_parameters(
+  struct DDS_StringSeq * const cft_parameters,
+  const rcutils_string_array_t * const cft_expression_parameters)
+{
+  // Cache value locally to avoid conversion warnings on Windows.
+  const DDS_Long params_len = static_cast<DDS_Long>(cft_expression_parameters->size);
+
+  if (!DDS_StringSeq_ensure_length(cft_parameters, params_len, params_len)) {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to ensure length for cft parameters sequence")
+    return RMW_RET_ERROR;
+  }
+  if (!DDS_StringSeq_from_array(
+      cft_parameters,
+      const_cast<const char **>(cft_expression_parameters->data),
+      params_len))
+  {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to copy data for cft parameters sequence")
+    return RMW_RET_ERROR;
+  }
+  return RMW_RET_OK;
+}
+
 rmw_ret_t
 rmw_connextdds_create_contentfilteredtopic(
   rmw_context_impl_t * const ctx,
@@ -236,6 +354,7 @@ rmw_connextdds_create_contentfilteredtopic(
   DDS_Topic * const base_topic,
   const char * const cft_name,
   const char * const cft_filter,
+  const rcutils_string_array_t * const cft_expression_parameters,
   DDS_TopicDescription ** const cft_out)
 {
   UNUSED_ARG(ctx);
@@ -243,13 +362,28 @@ rmw_connextdds_create_contentfilteredtopic(
   RMW_CONNEXT_ASSERT(nullptr != cft_filter)
 
   struct DDS_StringSeq cft_parameters = DDS_SEQUENCE_INITIALIZER;
-  DDS_StringSeq_ensure_length(&cft_parameters, 0, 0);
+  auto scope_exit_cft_params = rcpputils::make_scope_exit(
+    [&cft_parameters]() {
+      if (!DDS_StringSeq_finalize(&cft_parameters)) {
+        RMW_CONNEXT_LOG_ERROR_SET("failed to finalize cft parameters sequence")
+      }
+    });
+  if (cft_expression_parameters) {
+    if (RMW_RET_OK !=
+      rmw_connextdds_initialize_cft_parameters(&cft_parameters, cft_expression_parameters))
+    {
+      RMW_CONNEXT_LOG_ERROR_SET("failed to rmw_connextdds_initialize_cft_parameters")
+      return RMW_RET_ERROR;
+    }
+  }
 
   *cft_out = nullptr;
 
   DDS_ContentFilteredTopic * cft_topic =
-    DDS_DomainParticipant_create_contentfilteredtopic(
-    dp, cft_name, base_topic, cft_filter, &cft_parameters);
+    DDS_DomainParticipant_create_contentfilteredtopic_with_filter(
+    dp, cft_name, base_topic, cft_filter, &cft_parameters,
+    rti_connext_dds_custom_sql_filter::PLUGIN_NAME);
+
   if (nullptr == cft_topic) {
     RMW_CONNEXT_LOG_ERROR_A_SET(
       "failed to create content-filtered topic: "
@@ -354,6 +488,7 @@ rmw_connextdds_get_datawriter_qos(
         // TODO(asorbini) this value is not actually used, remove it
         &qos->publish_mode,
         &qos->lifespan,
+        &qos->user_data,
         qos_policies,
         pub_options,
         nullptr /* sub_options */))
@@ -366,6 +501,32 @@ rmw_connextdds_get_datawriter_qos(
     qos->publish_mode.kind = DDS_ASYNCHRONOUS_PUBLISH_MODE_QOS;
   }
 
+#if RMW_CONNEXT_DEFAULT_RELIABILITY_OPTIMIZATIONS
+  // The default settings for the RTPS reliability protocol are not very
+  // responsive, and they cause some unit tests to fail. These optimizations
+  // have been derived from profile `Optimization.ReliabilityProtocol.Common`
+  // available in Connext 6+. `Generic.StrictReliable` is the equivalent
+  // profile in 5.3.1. Changes are limited to `DDS_RtpsReliableWriterProtocol_t`.
+  if (ctx->optimize_reliability) {
+    // All write() calls will block (for at most max_blocking_time) once the send_window
+    // is filled with samples that haven't yet been acknowledged by all active readers.
+    qos->protocol.rtps_reliable_writer.min_send_window_size = 40;
+    qos->protocol.rtps_reliable_writer.max_send_window_size = 40;  // fixed size window
+    qos->protocol.rtps_reliable_writer.heartbeats_per_max_samples = 10;  // 1 every 4
+    qos->protocol.rtps_reliable_writer.heartbeat_period = {0, 200000000};  // 200ms
+    qos->protocol.rtps_reliable_writer.late_joiner_heartbeat_period = {0, 20000000};  // 20ms
+    qos->protocol.rtps_reliable_writer.fast_heartbeat_period = {0, 20000000};  // 20ms
+    qos->protocol.rtps_reliable_writer.max_heartbeat_retries = 500;  // 10s @ 50hz
+    // Force the writer to reply immediately to ACKNACK's received from a writer.
+    qos->protocol.rtps_reliable_writer.max_nack_response_delay = DDS_DURATION_ZERO;
+    // When the number of unack'd samples reaches the high_watermark the fast_heartbeat_period
+    // is used. When the number dips below the low_watermark, the heartbeat_period is used.
+    // These numbers are tied to the send_window size.
+    qos->protocol.rtps_reliable_writer.high_watermark = 25;
+    qos->protocol.rtps_reliable_writer.low_watermark = 10;
+  }
+#endif /* RMW_CONNEXT_DEFAULT_RELIABILITY_OPTIMIZATIONS */
+
 #if RMW_CONNEXT_DEFAULT_LARGE_DATA_OPTIMIZATIONS
   // Unless disabled, optimize the DataWriter's reliability protocol to
   // better handle large data samples. These are *bounded* types whose
@@ -443,6 +604,7 @@ rmw_connextdds_get_datareader_qos(
         &qos->resource_limits,
         nullptr /* publish_mode */,
         nullptr /* Lifespan is a writer-only qos policy */,
+        &qos->user_data,
         qos_policies,
         nullptr /* pub_options */,
         sub_options))
@@ -451,6 +613,17 @@ rmw_connextdds_get_datareader_qos(
     }
   }
 
+#if RMW_CONNEXT_DEFAULT_RELIABILITY_OPTIMIZATIONS
+  // The default settings for the RTPS reliability protocol are not very
+  // responsive, and they cause some unit tests to fail. These optimizations
+  // are dual to those applied in rmw_connextdds_get_datawriter_qos().
+  // Changes are limited to `DDS_RtpsReliableReaderProtocol_t`.
+  if (ctx->optimize_reliability) {
+    qos->protocol.rtps_reliable_reader.min_heartbeat_response_delay = DDS_DURATION_ZERO;
+    qos->protocol.rtps_reliable_reader.max_heartbeat_response_delay = DDS_DURATION_ZERO;
+  }
+#endif /* RMW_CONNEXT_DEFAULT_RELIABILITY_OPTIMIZATIONS */
+
 #if RMW_CONNEXT_DEFAULT_LARGE_DATA_OPTIMIZATIONS
   // Unless disabled, optimize the DataReader's reliability protocol to
   // better handle large data samples. These are *bounded* types whose
@@ -570,7 +743,6 @@ rmw_connextdds_write_message(
       // enable WriteParams::replace_auto to retrieve SN of published message
       write_params.replace_auto = DDS_BOOLEAN_TRUE;
     }
-
     if (DDS_RETCODE_OK !=
       DDS_DataWriter_write_w_params_untypedI(
         pub->writer(), message, &write_params))
@@ -640,9 +812,9 @@ rmw_connextdds_take_samples(
     RMW_CONNEXT_LOG_ERROR_SET("failed to take data from DDS reader")
     return RMW_RET_ERROR;
   }
-  RMW_CONNEXT_ASSERT(data_len > 0)(void) RMW_Connext_Uint8ArrayPtrSeq_loan_contiguous(
+  RMW_CONNEXT_ASSERT(data_len > 0)(void) RMW_Connext_MessagePtrSeq_loan_contiguous(
     sub->data_seq(),
-    reinterpret_cast<rcutils_uint8_array_t **>(data_buffer),
+    reinterpret_cast<RMW_Connext_Message **>(data_buffer),
     data_len,
     data_len);
 
@@ -654,11 +826,11 @@ rmw_connextdds_return_samples(
   RMW_Connext_Subscriber * const sub)
 {
   void ** data_buffer = reinterpret_cast<void **>(
-    RMW_Connext_Uint8ArrayPtrSeq_get_contiguous_buffer(sub->data_seq()));
+    RMW_Connext_MessagePtrSeq_get_contiguous_buffer(sub->data_seq()));
   const DDS_Long data_len =
-    RMW_Connext_Uint8ArrayPtrSeq_get_length(sub->data_seq());
+    RMW_Connext_MessagePtrSeq_get_length(sub->data_seq());
 
-  if (!RMW_Connext_Uint8ArrayPtrSeq_unloan(sub->data_seq())) {
+  if (!RMW_Connext_MessagePtrSeq_unloan(sub->data_seq())) {
     RMW_CONNEXT_LOG_ERROR_SET("failed to unloan sample sequence")
     return RMW_RET_ERROR;
   }
@@ -675,6 +847,62 @@ rmw_connextdds_return_samples(
   return RMW_RET_OK;
 }
 
+rmw_ret_t
+rmw_connextdds_count_unread_samples(
+  RMW_Connext_Subscriber * const sub,
+  size_t & unread_count)
+{
+  DDS_Boolean is_loan = DDS_BOOLEAN_TRUE;
+  DDS_Long data_len = 0;
+  void ** data_buffer = nullptr;
+  DDS_SampleInfoSeq info_seq = DDS_SEQUENCE_INITIALIZER;
+
+  unread_count = 0;
+  DDS_ReturnCode_t rc = DDS_RETCODE_ERROR;
+  do {
+    rc = DDS_DataReader_read_or_take_instance_untypedI(
+      sub->reader(),
+      &is_loan,
+      &data_buffer,
+      &data_len,
+      &info_seq,
+      0 /* data_seq_len */,
+      0 /* data_seq_max_len */,
+      DDS_BOOLEAN_TRUE /* data_seq_has_ownership */,
+      NULL /* data_seq_contiguous_buffer_for_copy */,
+      1 /* data_size -- ignored because loaning*/,
+      DDS_LENGTH_UNLIMITED /* max_samples */,
+      &DDS_HANDLE_NIL /* a_handle */,
+  #if !RMW_CONNEXT_DDS_API_PRO_LEGACY
+      NULL /* topic_query_guid */,
+  #endif /* RMW_CONNEXT_DDS_API_PRO_LEGACY */
+      DDS_NOT_READ_SAMPLE_STATE,
+      DDS_ANY_VIEW_STATE,
+      DDS_ANY_INSTANCE_STATE,
+      DDS_BOOLEAN_FALSE /* take */);
+    if (DDS_RETCODE_NO_DATA == rc) {
+      continue;
+    }
+    if (DDS_RETCODE_OK != rc && DDS_RETCODE_NO_DATA != rc) {
+      RMW_CONNEXT_LOG_ERROR_SET("failed to read data from DDS reader")
+      return RMW_RET_ERROR;
+    }
+    if (DDS_RETCODE_OK == rc) {
+      unread_count += data_len;
+      rc = DDS_DataReader_return_loan_untypedI(
+        sub->reader(),
+        data_buffer,
+        data_len,
+        &info_seq);
+      if (DDS_RETCODE_OK != rc) {
+        RMW_CONNEXT_LOG_ERROR_SET("failed to return loan to DDS reader")
+        return RMW_RET_ERROR;
+      }
+    }
+  } while (rc == DDS_RETCODE_OK);
+
+  return RMW_RET_OK;
+}
 
 rmw_ret_t
 rmw_connextdds_filter_sample(
@@ -1070,6 +1298,7 @@ rmw_connextdds_dcps_publication_on_data(rmw_context_impl_t * const ctx)
         &dp_guid,
         data->topic_name,
         data->type_name,
+        &data->user_data,
         &data->reliability,
         &data->durability,
         &data->deadline,
@@ -1153,6 +1382,7 @@ rmw_connextdds_dcps_subscription_on_data(rmw_context_impl_t * const ctx)
         &dp_guid,
         data->topic_name,
         data->type_name,
+        &data->user_data,
         &data->reliability,
         &data->durability,
         &data->deadline,
@@ -1243,3 +1473,98 @@ rmw_connextdds_enable_security(
 
   return RMW_RET_OK;
 }
+
+
+rmw_ret_t
+rmw_connextdds_set_cft_filter_expression(
+  DDS_TopicDescription * const topic_desc,
+  const char * const cft_expression,
+  const rcutils_string_array_t * const cft_expression_parameters)
+{
+  DDS_ContentFilteredTopic * const cft_topic =
+    DDS_ContentFilteredTopic_narrow(topic_desc);
+
+  struct DDS_StringSeq cft_parameters = DDS_SEQUENCE_INITIALIZER;
+  auto scope_exit_cft_parameters = rcpputils::make_scope_exit(
+    [&cft_parameters]() {
+      if (!DDS_StringSeq_finalize(&cft_parameters)) {
+        RMW_CONNEXT_LOG_ERROR_SET("failed to finalize cft parameters sequence")
+      }
+    });
+  if (nullptr != cft_expression_parameters) {
+    if (RMW_RET_OK !=
+      rmw_connextdds_initialize_cft_parameters(&cft_parameters, cft_expression_parameters))
+    {
+      RMW_CONNEXT_LOG_ERROR_SET("failed to rmw_connextdds_initialize_cft_parameters")
+      return RMW_RET_ERROR;
+    }
+  }
+
+  DDS_ReturnCode_t ret =
+    DDS_ContentFilteredTopic_set_expression(cft_topic, cft_expression, &cft_parameters);
+  if (DDS_RETCODE_OK != ret) {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to set content-filtered topic")
+    return RMW_RET_ERROR;
+  }
+  return RMW_RET_OK;
+}
+
+rmw_ret_t
+rmw_connextdds_get_cft_filter_expression(
+  DDS_TopicDescription * const topic_desc,
+  rcutils_allocator_t * const allocator,
+  rmw_subscription_content_filter_options_t * const options)
+{
+  DDS_ContentFilteredTopic * const cft_topic =
+    DDS_ContentFilteredTopic_narrow(topic_desc);
+
+  // get filter_expression
+  const char * filter_expression = DDS_ContentFilteredTopic_get_filter_expression(cft_topic);
+  if (!filter_expression) {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to get filter expression")
+    return RMW_RET_ERROR;
+  }
+
+  // get parameters
+  struct DDS_StringSeq parameters = DDS_SEQUENCE_INITIALIZER;
+  DDS_ReturnCode_t status =
+    DDS_ContentFilteredTopic_get_expression_parameters(cft_topic, &parameters);
+  if (DDS_RETCODE_OK != status) {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to get expression parameters")
+    return RMW_RET_ERROR;
+  }
+  auto scope_exit_parameters_delete =
+    rcpputils::make_scope_exit(
+    [&parameters]()
+    {
+      DDS_StringSeq_finalize(&parameters);
+    });
+
+  const DDS_Long parameters_len = DDS_StringSeq_get_length(&parameters);
+  std::vector<const char *> expression_parameters;
+  expression_parameters.reserve(parameters_len);
+
+  for (DDS_Long i = 0; i < parameters_len; ++i) {
+    const char * ref = *DDS_StringSeq_get_reference(&parameters, i);
+    if (!ref) {
+      RMW_CONNEXT_LOG_ERROR_A_SET(
+        "failed to get a reference for parameters with index %d", i)
+      return RMW_RET_ERROR;
+    }
+
+    expression_parameters.push_back(ref);
+  }
+
+  if (RMW_RET_OK != rmw_subscription_content_filter_options_init(
+      filter_expression,
+      expression_parameters.size(),
+      expression_parameters.data(),
+      allocator,
+      options))
+  {
+    RMW_CONNEXT_LOG_ERROR_SET("failed to rmw_subscription_content_filter_options_init")
+    return RMW_RET_ERROR;
+  }
+
+  return RMW_RET_OK;
+}
