@@ -649,6 +649,8 @@ rmw_connextdds_finalize_participant_factory_context(
   RMW_CONNEXT_ASSERT(nullptr != RMW_Connext_fv_FactoryContext)
   rmw_connextdds_api_micro * const ctx_api = RMW_Connext_fv_FactoryContext;
 
+  RMW_Connext_fv_FactoryContext = nullptr;
+
   RT_Registry_T * registry =
     DDS_DomainParticipantFactory_get_registry(
     RMW_Connext_gv_DomainParticipantFactory);
@@ -860,6 +862,17 @@ rmw_connextdds_initialize_participant_qos_impl(
   return RMW_RET_OK;
 }
 
+rmw_ret_t
+rmw_connextdds_configure_participant(
+  rmw_context_impl_t * const ctx,
+  DDS_DomainParticipant * const participant)
+{
+  UNUSED_ARG(participant);
+  UNUSED_ARG(ctx);
+
+  return RMW_RET_OK;
+}
+
 rmw_ret_t
 rmw_connextdds_create_contentfilteredtopic(
   rmw_context_impl_t * const ctx,
@@ -867,6 +880,7 @@ rmw_connextdds_create_contentfilteredtopic(
   DDS_Topic * const base_topic,
   const char * const cft_name,
   const char * const cft_filter,
+  const rcutils_string_array_t * const cft_expression_parameters,
   DDS_TopicDescription ** const cft_out)
 {
   UNUSED_ARG(ctx);
@@ -874,6 +888,7 @@ rmw_connextdds_create_contentfilteredtopic(
   UNUSED_ARG(base_topic);
   UNUSED_ARG(cft_name);
   UNUSED_ARG(cft_filter);
+  UNUSED_ARG(cft_expression_parameters);
   UNUSED_ARG(cft_out);
   return RMW_RET_UNSUPPORTED;
 }
@@ -1000,6 +1015,7 @@ rmw_connextdds_get_datawriter_qos(
       &qos->resource_limits,
       &qos->publish_mode,
       nullptr /* Micro doesn't support DDS_LifespanQosPolicy */,
+      &qos->user_data,
       qos_policies,
       pub_options,
       nullptr /* sub_options */))
@@ -1049,6 +1065,7 @@ rmw_connextdds_get_datareader_qos(
       &qos->resource_limits,
       nullptr /* publish_mode */,
       nullptr /* Lifespan is a writer-only qos policy */,
+      &qos->user_data,
       qos_policies,
       nullptr /* pub_options */,
       sub_options))
@@ -1188,6 +1205,40 @@ rmw_connextdds_return_samples(
   return RMW_RET_OK;
 }
 
+rmw_ret_t
+rmw_connextdds_count_unread_samples(
+  RMW_Connext_Subscriber * const sub,
+  size_t & unread_count)
+{
+  unread_count = 0;
+  DDS_SampleInfoSeq info_seq = DDS_SEQUENCE_INITIALIZER;
+  DDS_UntypedSampleSeq data_seq = DDS_SEQUENCE_INITIALIZER;
+  DDS_ReturnCode_t rc = DDS_RETCODE_ERROR;
+  do {
+    rc = DDS_DataReader_read(
+      sub->reader(),
+      &data_seq,
+      &info_seq,
+      DDS_LENGTH_UNLIMITED,
+      DDS_NOT_READ_SAMPLE_STATE,
+      DDS_ANY_VIEW_STATE,
+      DDS_ANY_INSTANCE_STATE);
+    if (DDS_RETCODE_OK != rc && DDS_RETCODE_NO_DATA != rc) {
+      RMW_CONNEXT_LOG_ERROR_SET("failed to read data from DDS reader")
+      return RMW_RET_ERROR;
+    }
+    if (DDS_RETCODE_OK == rc) {
+      unread_count += DDS_UntypedSampleSeq_get_length(&data_seq);
+      rc = DDS_DataReader_return_loan(sub->reader(), &data_seq, &info_seq);
+      if (DDS_RETCODE_OK != rc) {
+        RMW_CONNEXT_LOG_ERROR_SET("failed to return loan to DDS reader")
+        return RMW_RET_ERROR;
+      }
+    }
+  } while (rc == DDS_RETCODE_OK);
+  return RMW_RET_OK;
+}
+
 rmw_ret_t
 rmw_connextdds_filter_sample(
   RMW_Connext_Subscriber * const sub,
@@ -1751,6 +1802,7 @@ rmw_connextdds_dcps_publication_on_data(rmw_context_impl_t * const ctx)
         &dp_guid,
         qdata.data.topic_name,
         qdata.data.type_name,
+        &data->user_data,
         &qdata.data.reliability,
         &qdata.data.durability,
         &qdata.data.deadline,
@@ -1809,6 +1861,7 @@ rmw_connextdds_dcps_subscription_on_data(rmw_context_impl_t * const ctx)
         &dp_guid,
         qdata.data.topic_name,
         qdata.data.type_name,
+        &data->user_data,
         &qdata.data.reliability,
         &qdata.data.durability,
         &qdata.data.deadline,
@@ -1952,3 +2005,27 @@ rmw_connextdds_enable_security(
   return RMW_RET_ERROR;
 #endif /* RMW_CONNEXT_ENABLE_SECURITY */
 }
+
+rmw_ret_t
+rmw_connextdds_set_cft_filter_expression(
+  DDS_TopicDescription * const topic_desc,
+  const char * const cft_expression,
+  const rcutils_string_array_t * const cft_expression_parameters)
+{
+  UNUSED_ARG(topic_desc);
+  UNUSED_ARG(cft_expression);
+  UNUSED_ARG(cft_expression_parameters);
+  return RMW_RET_UNSUPPORTED;
+}
+
+rmw_ret_t
+rmw_connextdds_get_cft_filter_expression(
+  DDS_TopicDescription * const topic_desc,
+  rcutils_allocator_t * const allocator,
+  rmw_subscription_content_filter_options_t * const options)
+{
+  UNUSED_ARG(topic_desc);
+  UNUSED_ARG(allocator);
+  UNUSED_ARG(options);
+  return RMW_RET_UNSUPPORTED;
+}
